package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.*
import com.mojang.math.Matrix4f
import net.minecraft.client.Minecraft
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.client.renderer.MultiBufferSource
import net.minecraft.client.renderer.blockentity.BlockEntityRenderer
import net.minecraft.client.renderer.blockentity.BlockEntityRendererProvider
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.phys.Vec3
import org.lwjgl.opengl.GL30
import ru.dbotthepony.mc.otm.Registry
import ru.dbotthepony.mc.otm.block.entity.blackhole.BlockEntityBlackHole
import ru.dbotthepony.mc.otm.core.*

private const val BEAM_WIDTH = 0.2

private fun pushQuad(matrix4f: Matrix4f, builder: BufferBuilder, quad: Array<Vector>, fwdRotation: Double, rightRotation: Double, topRotation: Double, anim: Float) {
	var a = quad[0]
	var b = quad[1]
	var c = quad[2]
	var d = quad[3]

	if (fwdRotation != 0.0) {
		a = a.rotateAroundAxis(VECTOR_FORWARD, fwdRotation)
		b = b.rotateAroundAxis(VECTOR_FORWARD, fwdRotation)
		c = c.rotateAroundAxis(VECTOR_FORWARD, fwdRotation)
		d = d.rotateAroundAxis(VECTOR_FORWARD, fwdRotation)
	}

	if (rightRotation != 0.0) {
		a = a.rotateAroundAxis(VECTOR_RIGHT, rightRotation)
		b = b.rotateAroundAxis(VECTOR_RIGHT, rightRotation)
		c = c.rotateAroundAxis(VECTOR_RIGHT, rightRotation)
		d = d.rotateAroundAxis(VECTOR_RIGHT, rightRotation)
	} else if (topRotation != 0.0) {
		a = a.rotateAroundAxis(VECTOR_UP, topRotation)
		b = b.rotateAroundAxis(VECTOR_UP, topRotation)
		c = c.rotateAroundAxis(VECTOR_UP, topRotation)
		d = d.rotateAroundAxis(VECTOR_UP, topRotation)
	}

	builder.vertex(matrix4f, b.x.toFloat(), b.y.toFloat(), b.z.toFloat()).uv(0f, anim).endVertex()
	builder.vertex(matrix4f, c.x.toFloat(), c.y.toFloat(), c.z.toFloat()).uv(1f, anim).endVertex()
	builder.vertex(matrix4f, d.x.toFloat(), d.y.toFloat(), d.z.toFloat()).uv(1f, anim + 64f).endVertex()
	builder.vertex(matrix4f, a.x.toFloat(), a.y.toFloat(), a.z.toFloat()).uv(0f, anim + 64f).endVertex()
}

private val quadA = quadY(-BEAM_WIDTH)
private val quadB = quadY(BEAM_WIDTH)
private val quadC = quadZ(-BEAM_WIDTH)
private val quadD = quadZ(BEAM_WIDTH)

private fun beam(matrix4f: Matrix4f, builder: BufferBuilder, fwdRotation: Double, rightRotation: Double, topRotation: Double, anim: Float) {
	pushQuad(matrix4f, builder, quadA, fwdRotation, rightRotation, topRotation, anim)
	pushQuad(matrix4f, builder, quadB, fwdRotation, rightRotation, topRotation, anim)
	pushQuad(matrix4f, builder, quadC, fwdRotation, rightRotation, topRotation, anim)
	pushQuad(matrix4f, builder, quadD, fwdRotation, rightRotation, topRotation, anim)
}

private fun quadY(z: Double): Array<Vector> {
	return arrayOf(
		Vector(0.0, -BEAM_WIDTH, z),
		Vector(64.0, -BEAM_WIDTH, z),
		Vector(64.0, BEAM_WIDTH, z),
		Vector(0.0, BEAM_WIDTH, z)
	)
}

private fun quadZ(y: Double): Array<Vector> {
	return arrayOf(
		Vector(0.0, y, -BEAM_WIDTH),
		Vector(64.0, y, -BEAM_WIDTH),
		Vector(64.0, y, BEAM_WIDTH),
		Vector(0.0, y, BEAM_WIDTH)
	)
}

private val BEACON_BEAM = ResourceLocation("minecraft", "textures/entity/beacon_beam.png")

class BlackHoleRenderer(private val context: BlockEntityRendererProvider.Context) : BlockEntityRenderer<BlockEntityBlackHole> {
	override fun render(
		blockEntityBlackHole: BlockEntityBlackHole,
		v: Float,
		poseStack: PoseStack,
		multiBufferSource: MultiBufferSource,
		i: Int,
		i1: Int
	) {
		RenderHelper.setDrawColor(RGBAColor.BLACK)
		RenderSystem.setShader(GameRenderer::getPositionColorShader)

		RenderSystem.depthFunc(GL30.GL_LESS)
		RenderSystem.depthMask(true)
		RenderSystem.enableDepthTest()
		RenderSystem.disableCull()

		poseStack.pushPose()
		poseStack.translate(0.5, -blockEntityBlackHole.gravitationStrength / 2 + 0.5, 0.5)
		RenderHelper.colorSphere(poseStack, blockEntityBlackHole.gravitationStrength.toFloat())

		RenderSystem.enableCull()
		RenderSystem.enableTexture()

		poseStack.popPose()

		val ply = Minecraft.getInstance().player

		if (ply != null && ply.mainHandItem.item == Registry.Items.GRAVITATION_STABILIZER) {
			RenderSystem.setShader(GameRenderer::getPositionTexShader)
			RenderSystem.setShaderTexture(0, BEACON_BEAM)
			RenderSystem.disableCull()

			val builder = Tesselator.getInstance().builder
			builder.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION_TEX)

			val anim = (((System.currentTimeMillis().toDouble() / 100.0) % 100.0).toFloat() / 100f) * 64f
			val rotation = System.currentTimeMillis().toDouble() / 1000.0

			poseStack.pushPose()
			poseStack.translate(0.5, 0.5, 0.5)
			val matrix = poseStack.last().pose()

			beam(matrix, builder, rotation, 0.0, 0.0, anim)
			beam(matrix, builder, rotation, 0.0, Math.PI / 2, anim)
			beam(matrix, builder, rotation, 0.0, -Math.PI / 2, anim)
			beam(matrix, builder, rotation, 0.0, Math.PI, anim)
			beam(matrix, builder, rotation, Math.PI / 2, 0.0, anim)
			beam(matrix, builder, rotation, -Math.PI / 2, 0.0, anim)
			beam(matrix, builder, rotation, Math.PI, 0.0, anim)

			poseStack.popPose()

			builder.end()
			BufferUploader.end(builder)
			RenderSystem.enableCull()
		}
	}

	override fun shouldRenderOffScreen(p_112306_: BlockEntityBlackHole): Boolean {
		return true
	}

	override fun shouldRender(p_173568_: BlockEntityBlackHole, p_173569_: Vec3): Boolean {
		return true
	}

	override fun getViewDistance(): Int {
		return 512
	}
}